/*
 * Velbus Serial Communications Library for .NET
 * Copyright (C) 2007 Velleman
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 * 
 */
#region "Imports"

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Threading;

#endregion 

namespace Velleman.Velbus
{
    /// <summary>
    /// The dispatcher class is in charge of sending packets over the bus
    /// in an asynchronous way, while respecting a fixed delay between each
    /// packet.
    /// </summary>
    public class Dispatcher
    {
        #region "Public Fields"

        /// <summary>
        /// Initial queue size.
        /// </summary>
        public const int QUEUE_SIZE = 500;

        #endregion

        /// <summary>
        /// Bus interface that this class is attached to.
        /// </summary>
        private IBusInterface m_Bus;

        /// <summary>
        /// Internal dispatch queue.
        /// </summary>
        private Queue m_DispatchQueue;

        /// <summary>
        /// Main worker thread in charge of dispatching queued packets.
        /// </summary>
        private Thread m_WorkerThread = null;

        /// <summary>
        /// Used to signal the worker thread that it should finish.
        /// </summary>
        private volatile bool m_Terminated = false;

        /// <summary>
        /// Signal used to indicate that a packet is waiting in the queue.
        /// </summary>
        private ManualResetEvent m_EventPacketQueued;

        #region "Constructors"

        /// <summary>
        /// Initializes a new instance of the Dispatcher class attached to the
        /// specified SerialBus object.
        /// </summary>
        /// <param name="serialBus">The SerialBus to attach to.</param>
        public Dispatcher(SerialBus serialBus)
        {
            m_Bus = serialBus;
            m_DispatchQueue = new Queue(Dispatcher.QUEUE_SIZE);

            m_EventPacketQueued = new ManualResetEvent(false);
        }

        #endregion

        #region "Private Methods"

        /// <summary>
        /// Main loop for the worker thread.
        /// </summary>
        private void DispatchingHandler()
        {
            while (!m_Terminated)
            {
                if (WaitHandle.WaitAll(new WaitHandle[] { m_EventPacketQueued }, 100, true))
                {
                    m_EventPacketQueued.Reset();

                    while (!m_Terminated)
                    {
                        lock(m_DispatchQueue)
                        {
                            if (m_DispatchQueue.Count <= 0)
                            {
                                break;
                            }
                            else
                            {
                                lock (m_Bus)
                                {
                                    Packet packet = (Packet)m_DispatchQueue.Dequeue();
                                    m_Bus.SendBlocking(packet);
                                }
                            }
                        }
                        Thread.Sleep(m_Bus.Delay); // dispatching
                    }
                }               
            }
        }

        #endregion

        #region "Public Methods"

        /// <summary>
        /// Adds a packet to the dispatching queue.
        /// </summary>
        /// <param name="packet">The packet to add.</param>
        public void Dispatch(Packet packet)
        {
            lock (m_DispatchQueue)
            {
                m_DispatchQueue.Enqueue(packet.Clone());
            }

            m_EventPacketQueued.Set();
        }

        /// <summary>
        /// Starts the dispatcher. This will create a worker thread in charge
        /// of processing the queue whenever a packets are waiting.
        /// </summary>
        public void Start()
        {
            if (m_WorkerThread == null)
            {
                m_Terminated = false;
                m_EventPacketQueued.Reset();

                m_WorkerThread = new Thread(new ThreadStart(this.DispatchingHandler));
                m_WorkerThread.Start();
            }
        }

        /// <summary>
        /// Stops the dispatcher.
        /// </summary>
        public void Stop()
        {
            if (m_WorkerThread != null)
            {
                m_Terminated = true;
                m_WorkerThread.Join();

                m_DispatchQueue.Clear();

                m_WorkerThread = null;
            }
        }

        #endregion
    }
}