/*
 * Velbus Serial Communications Library for .NET
 * Copyright (C) 2007 Velleman
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 * 
 */
#region "Imports"

using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.IO.Ports;
using System.Drawing;

#endregion

namespace Velleman.Velbus
{
    /// <summary>
    /// The SerialBus class is serial communications wrapper that operates
    /// with Velbus packets instead of raw data.
    /// </summary>
    [ToolboxBitmapAttribute(typeof(SerialBus), "images.SerialBus.bmp")] 
    public partial class SerialBus : Component, IBusInterface
    {
        #region "Public Fields"

        /// <summary>
        /// Default delay between each packet sent in non-blocking (default) mode.
        /// </summary>
        public const int DEFAULT_DELAY = 50;

        #endregion

        /// <summary>
        /// Underlying SerialPort object. This class acts as a wrapper around
        /// this object to convert its raw bytestream into packets used by
        /// the Velbus protocol.
        /// </summary>
        private SerialPort m_SerialPort;

        /// <summary>
        /// The dispatcher is responsable for sending each packet over the serial bus
        /// while respecting a delay between each packet. This delay is necessary to
        /// prevent congestion of the Velbus network.
        /// </summary>
        private Dispatcher m_Dispatcher;

        /// <summary>
        /// Parser that understands the Velbus protocol. The parser is used to convert
        /// raw incoming data into Velbus packets.
        /// </summary>
        private PacketParser m_Parser;

        /// <summary>
        /// Intermediate input buffer.
        /// </summary>
        private byte[] m_InBuffer = new byte[512];

        /// <summary>
        /// Delay between asynchronous operations
        /// </summary>
        private volatile int m_Delay = DEFAULT_DELAY;

        #region "Constructors"

        /// <summary>
        /// Initializes a new instance of the SerialBus class.
        /// </summary>
        public SerialBus()
            : this("COM1")
        {
            InitializeComponent();
        }

        /// <summary>
        /// Initializes a new instance of the SerialBus class using
        /// the specified IContainer object. 
        /// </summary>
        /// <param name="container">An interface to a container.</param>
        public SerialBus(IContainer container)
            : this("COM1")
        {
            container.Add(this);

            InitializeComponent();
        }

        /// <summary>
        /// Initializes a new instance of the SerialPort class using
        /// the specified port name. 
        /// </summary>
        /// <param name="portName">The port to use (for example, COM1).</param>
        public SerialBus(String portName) : this(portName, 38400) { }

        /// <summary>
        /// Initializes a new instance of the SerialPort class using
        /// the specified port name and baud rate.
        /// </summary>
        /// <param name="portName">The port to use (for example, COM1).</param>
        /// <param name="baudRate">The baud rate.</param>
        public SerialBus(String portName, int baudRate)
        {
            m_SerialPort = new SerialPort(portName, baudRate);
            m_SerialPort.Parity = Parity.None;
            m_SerialPort.StopBits = StopBits.One;
            m_SerialPort.DataBits = 8;
            m_SerialPort.DtrEnable = false;
            m_SerialPort.RtsEnable = true;
            m_SerialPort.DataReceived += new SerialDataReceivedEventHandler(SerialDataReceivedHandler);
            m_SerialPort.ReceivedBytesThreshold = 1;
            m_Dispatcher = new Dispatcher(this);

            m_Parser = new PacketParser();
        }

        #endregion

        #region "Private Methods"

        /// <summary>
        /// Event handler for the DataReceived event of underlying SerialPort
        /// component. This method feeds the incoming data into the parser and
        /// notifies the user if the stream contains any packets.
        /// </summary>
        /// <param name="source">This will always be the SerialPort component.</param>
        /// <param name="args">Event arguments.</param>
        private void SerialDataReceivedHandler(object source, SerialDataReceivedEventArgs args)
        {
            int BytesRead = m_SerialPort.Read(m_InBuffer, 0, m_SerialPort.BytesToRead);
            m_Parser.Feed(m_InBuffer, BytesRead);

            Packet packet = m_Parser.Next();
            while (packet != null)
            {
                if (PacketReceived != null)
                    PacketReceived(this, new BusPacketIoEventArgs(packet));
                packet = m_Parser.Next();
            }
        }

        #endregion

        #region "Public Events"

        ///<summary>
        /// Represents the method that will handle the packet sent event
        /// of a SerialBus object.
        /// </summary>
        public event BusPacketIoEventHandler PacketReceived;

        /// <summary>
        /// Represents the method that will handle the packet received event
        /// of a SerialBus object.
        /// </summary>
        public event BusPacketIoEventHandler PacketSent;

        #endregion

        #region "Public Properties"

        /// <summary>
        /// Gets or sets the delay (in ms) between each packet sent over
        /// the bus. This is necessary because the serial port can send out
        /// packets much faster than the Velbus modules can process them.
        /// A delay of 50 ms has proven to be sufficient.
        /// </summary>
        public int Delay
        {
            get
            {
                return m_Delay;
            }
            set
            {
                m_Delay = value;
            }
        }

        /// <summary>
        /// Gets or sets the port for communications, including but not limited
        /// to all available COM ports.
        /// </summary>
        public String PortName
        {
            get
            {
                return m_SerialPort.PortName;
            }
            set
            {
                m_SerialPort.PortName = value;
            }
        }

        /// <summary>
        /// Gets or sets the serial baud rate.
        /// </summary>
        public int BaudRate
        {
            get
            {
                return m_SerialPort.BaudRate;
            }
            set
            {
                m_SerialPort.BaudRate = value;
            }
        }

        #endregion

        #region "Public Methods"

        /// <summary>
        /// Opens a new serial connection to the Velbus network.
        /// </summary>
        public void Open()
        {
            if (!IsOpen)
            {
                m_SerialPort.Open();
                m_Dispatcher.Start();
            }
        }

        /// <summary>
        /// Closes the port connection, sets the IsOpen property to false, and
        /// clears the internal packet queue.
        /// </summary>
        public void Close()
        {
            if (IsOpen)
            {
                m_Dispatcher.Stop();
                m_SerialPort.Close();
            }
        }

        /// <summary>
        /// Gets the open or closed status of the connection.
        /// </summary>
        public bool IsOpen
        {
            get
            {
                return m_SerialPort.IsOpen;
            }
        }

        /// <summary>
        /// Sends a packet over the bus. In case of heavy traffic, the packet
        /// will get queued until it can be sent safely without overloading the
        /// Velbus network.
        /// </summary>
        /// <param name="packet">Velbus protocol packet to send.</param>
        public void Send(Packet packet)
        {
            /* Add the packet to the dispatching queue */
            m_Dispatcher.Dispatch(packet);
        }

        /// <summary>
        /// Sends a packet over the bus in blocking mode. This will not
        /// keep track of the necessary delay between each packet and might
        /// congest the Velbus network.
        /// </summary>
        /// <param name="packet">Velbus protocol packet to send</param>
        public void SendBlocking(Packet packet)
        {
            lock (m_SerialPort)
            {
                m_SerialPort.Write(packet.Pack(), 0, packet.Size);
                m_SerialPort.BaseStream.Flush();
            }

            /* Notify the user */
            if (PacketSent != null)
                PacketSent(this, new BusPacketIoEventArgs(packet));
        }

        #endregion
    }
}
